#!/bin/sh
TARFILE=$1
UPDLOG=/usr/kingfisher/var/update.txt
FWUGLOG=/usr/kingfisher/var/last_upgrade.txt
USRTMPDIR=/usr/tmp/update
ZIMAGE=$USRTMPDIR/zImage
RAMDISK=$USRTMPDIR/ramdisk.gz
REDBOOT=$USRTMPDIR/redboot.bin
USRIMG=/tmp/usr.img
MTD2_OFFSET=0
REDBOOT_BASE=0x60000000
REDBOOT_SIZE=0x60000
ZIMAGE_BASE=0x60060000
ZIMAGE_SIZE=0x200000
RAMDISK_BASE=0x60260000
RAMDISK_SIZE=0x280000
USRIMG_BASE=0x604e0000
USRIMG_SIZE=0xb00000

DIRBUF=/tmp/dirbuf
SAVEFILE=save1.tgz

MTDFLASH=/sbin/mtdflash
FISDIR=/sbin/fisdir

errexit ()
{
    echo $1 >$UPDLOG
    logger -t postinst $1
    if [ ! -e /tmp/.update.debug ]; then
        rm -rf $USRTMPDIR
        rm -f $USRIMG $DIRBUF /tmp/update
    fi
    # updater will not reboot - allow user to retry upgrade
    exit 9
}

# when run from a non-interactive ssh shell, default path is only /bin:/usr/bin
export PATH=/sbin:$PATH

mkdir -p /usr/kingfisher/var

# delete original package and copy usr image and saved data back to /tmp
rm $TARFILE
mv $USRTMPDIR/usr.img /tmp 2>/dev/null
mv $USRTMPDIR/$SAVEFILE /tmp 2>/dev/null

if [ -e $ZIMAGE -o -e $RAMDISK -o -e $REDBOOT ]; then
    # at least one flash partition needs to be updated
    if [ "`$MTDFLASH info /dev/mtd4 | grep 'flags = 0x4'`" ]; then
        # FIS dir partition is read only
        errexit "941 Flash directory is read only"
    fi
    # read FIS directory into a file
    if !($FISDIR -o $DIRBUF); then
        errexit "949 Cannot read directory"
    fi

    if [ -e $ZIMAGE ]; then
        # zImage (Linux kernel) is present in archive
        echo "141 Checking Linux kernel" > $UPDLOG
        if !($FISDIR -c $ZIMAGE -p $DIRBUF zImage && $MTDFLASH verify /dev/mtd1 $ZIMAGE); then
            # zImage is different to existing, so it needs to be updated
            echo "142 Updating Linux kernel" > $UPDLOG
            #  calculate length/crc and generate updated directory image. Set length to 2MB
            if !($FISDIR -u $ZIMAGE -p $DIRBUF -o $DIRBUF -b $ZIMAGE_BASE -l $ZIMAGE_SIZE zImage); then
                errexit "941 Cannot adjust zImage directory entry"
            fi
            # check whether Linux image will fit in flash partition
            MTD1_SIZE=$($FISDIR -d zImage | awk '/^[0-9]/ {print $2}')
            if [ $MTD1_SIZE -eq 2097152 ]; then
                # Linux will fit in mtd1 partition
                # Write zImage to /dev/mtd1
                if !($MTDFLASH erase /dev/mtd1 &&
                     $MTDFLASH write /dev/mtd1 $ZIMAGE &&
                     $MTDFLASH verify /dev/mtd1 $ZIMAGE); then
                    errexit "942 Cannot write Linux kernel to flash (1)"
                fi
            elif [ $MTD1_SIZE -eq 1048576 -a -e $RAMDISK ]; then
                # We have an old style (1MB) Linux partition
                # Write the first 1M of the Linux image to it, and write the rest to the start of the 
                #  ramdisk partition (/dev/mtd2). The ramdisk image will then be written at a 1M offset.
                echo "143 Updating Linux kernel - adjusting size" > $UPDLOG
                MTD2_OFFSET=0x100000
                # Write zImage part 1 to /dev/mtd1
                if !($MTDFLASH erase -l 0x100000 /dev/mtd1 &&
                     $MTDFLASH write -l 0x100000 /dev/mtd1 $ZIMAGE &&
                     $MTDFLASH verify -l 0x100000 /dev/mtd1 $ZIMAGE); then
                    errexit "943 Cannot write Linux kernel to flash (2)"
                fi
                # Write zImage part 2 to /dev/mtd2
                if !($MTDFLASH erase /dev/mtd2 &&
                     $MTDFLASH write -s 0x100000 /dev/mtd2 $ZIMAGE &&
                     $MTDFLASH verify -s 0x100000 /dev/mtd2 $ZIMAGE); then
                    errexit "944 Cannot write Linux kernel to flash (3)"
                fi
            else
                errexit "945 Invalid flash partition configuration ($MTD1_SIZE)"
            fi
        fi
    fi
                    
    if [ -e $RAMDISK ]; then
        # RAM disk image is present in archive
        echo "151 Checking RAM disk image" > $UPDLOG
        if !($FISDIR -c $RAMDISK -p $DIRBUF ramdisk && $MTDFLASH verify /dev/mtd2 $RAMDISK) || [ $MTD2_OFFSET -ne 0 ]; then
            # ramdisk image is different to existing, or zImage partition needs to be resized: need to update ramdisk
            echo "152 Updating RAM disk image" > $UPDLOG
            #  calculate length/crc and generate updated directory image
            if !($FISDIR -u $RAMDISK -p $DIRBUF -o $DIRBUF -b $RAMDISK_BASE -l $RAMDISK_SIZE ramdisk); then
                errexit "951 Cannot adjust RAM disk directory entry"
            fi
            # write ramdisk to /dev/mtd2, possibly offset by 1MB if zImage overflowed into ramdisk partition
            if !($MTDFLASH erase -o $MTD2_OFFSET /dev/mtd2 &&
                 $MTDFLASH write -o $MTD2_OFFSET /dev/mtd2 $RAMDISK &&
                 $MTDFLASH verify -o $MTD2_OFFSET /dev/mtd2 $RAMDISK); then
                errexit "952 Cannot write RAM disk image to flash"
            fi
        fi
    fi
                    
    if [ -e $REDBOOT ]; then
        # Redboot image is present in archive
        echo "153 Checking RedBoot" > $UPDLOG
        if !($FISDIR -c $REDBOOT -p $DIRBUF redboot && $MTDFLASH verify /dev/mtd0 $REDBOOT); then
            # redboot image is different to existing, so it needs to be updated
            echo "154 Updating RedBoot" > $UPDLOG
            #  calculate length/crc and generate updated directory image
            if !($FISDIR -u $REDBOOT -p $DIRBUF -o $DIRBUF -b $REDBOOT_BASE -l $REDBOOT_SIZE RedBoot); then
                errexit "953 Cannot adjust Redboot directory entry"
            fi
            # write redboot to /dev/mtd0
            if !($MTDFLASH erase /dev/mtd0 &&
                 $MTDFLASH write /dev/mtd0 $REDBOOT &&
                 $MTDFLASH verify /dev/mtd0 $REDBOOT); then
                errexit "954 Cannot write RedBoot image to flash"
            fi
        fi
    fi

    # update FIS directory in /dev/mtd4
    # Redboot is not interested in usr partition, so no need to update its size/CRC in directory
    echo "170 Updating flash directory" >$UPDLOG
    if !($MTDFLASH erase /dev/mtd4 &&
         $MTDFLASH write /dev/mtd4 $DIRBUF &&
         $MTDFLASH verify /dev/mtd4 $DIRBUF); then
        errexit "970 Cannot write FIS directory to flash"
    fi
fi                    

rm -rf $USRTMPDIR

if [ -e $USRIMG ]; then
    # /usr image is present in archive
    # syslog writes to flash, so stop it
    killall -q syslogd klogd
    sleep 2
    echo "161 Unmounting file system" > $UPDLOG
    waited=0
    while ! umount /usr && [ $waited -lt 30 ]; do
        # other ssh connections eg WinSCP may be holding a /usr folder open
        if [ $waited -ge 10 ]; then
            cp /usr/kingfisher/etc/dropbear_dss_host_key /etc/dropbear
            killall -q dropbear sshd
        fi
        sleep 2
        waited=$((waited+2))
    done
    if [ $waited -ge 30 ]; then
        errexit "961 File system is busy"
    fi

    if ! pidof dropbear; then
        /sbin/dropbear &
    fi
    # /usr is now in RAM disk - recreate status file
    mkdir -p /usr/kingfisher/var
    echo "162 Updating application" > $UPDLOG

    # write application to /dev/mtd3
    if !($MTDFLASH erase /dev/mtd3 &&
         $MTDFLASH write /dev/mtd3 $USRIMG &&
         $MTDFLASH verify /dev/mtd3 $USRIMG); then
        errexit "963 Cannot write application image to flash"
    fi
    # mount new image and restore saved settings
    echo "163 Remounting file system" > $UPDLOG
    if ! mount -t jffs2 mtd:usr /usr -o sync; then
        errexit "964 Cannot mount file system"
    fi
    # /usr is now back in flash
    mkdir -p /usr/kingfisher/var
    # restore settings (logs, config, MC31 indicator etc) that were backed up in preinstall.sh
    echo "165 Restoring settings" > $UPDLOG
    if ! tar -C/ -xzf /tmp/$SAVEFILE; then
        errexit "965 Cannot restore settings"
    fi
    # restart syslog
    if [ -x /etc/init.d/S01logging ]; then
        /etc/init.d/S01logging start
    else
        syslogd -O /usr/kingfisher/var/syslog00.txt -m0 -b4 -S -s64
        klogd
    fi
fi

echo "166 Synchronising file system" > $UPDLOG
sync
sleep 1
sync

logger -t postinst "Firmware upgrade completed" 2>/dev/null
echo "200 Upgrade complete" > $UPDLOG

# After TB+ has read the above completion code it will write something different to update.txt. Wait up to 10s for this to occur.
count=0; while (`grep -q 200 $UPDLOG` && [ $count -lt 10 ]); do sleep 1; count=$((count+1)); done

# updater will now reboot

exit 0


